/**
 *
 *  Copyright Openmind http://www.openmindonline.it
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package net.sourceforge.openutils.mgnlcriteria.jcr.query.xpath.impl;

import info.magnolia.cms.core.Content;
import info.magnolia.cms.core.DefaultContent;
import info.magnolia.cms.core.HierarchyManager;
import info.magnolia.cms.core.ItemType;
import info.magnolia.cms.core.Path;
import info.magnolia.cms.core.search.Query;
import info.magnolia.cms.core.search.QueryManager;
import info.magnolia.cms.core.search.QueryResult;
import info.magnolia.cms.security.AccessManager;
import info.magnolia.cms.security.Permission;
import info.magnolia.content2bean.Content2BeanException;
import info.magnolia.content2bean.Content2BeanUtil;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.jcr.Node;
import javax.jcr.NodeIterator;
import javax.jcr.RepositoryException;
import javax.jcr.query.InvalidQueryException;

import net.sourceforge.openutils.mgnlcriteria.jcr.query.JCRMagnoliaCriteriaQueryTranslator;
import net.sourceforge.openutils.mgnlcriteria.jcr.query.JCRQueryException;
import net.sourceforge.openutils.mgnlcriteria.jcr.query.xpath.XPathSelect;

import org.apache.commons.lang.StringUtils;


/**
 * Magnolia XPATH implementation of the <tt>Criteria</tt> interface which supports limiting the result set according to
 * the underlying JCR implementation (e.g. Jackrabbit 1.4+)
 * @author Federico Grilli
 * @version $Id: MagnoliaCriteriaWithLimitImpl.java 1481 2009-10-09 19:31:24Z fgrilli $
 */
public class MagnoliaCriteriaWithLimitImpl extends AbstractMagnoliaCriteriaImpl
{

    private static final String HIERARCHY_MANAGER = "hm";

    private static final String QUERY_MANAGER = "queryManager";

    public MagnoliaCriteriaWithLimitImpl(String path, QueryManager queryManager, String itemType)
    {
        super(path, queryManager, itemType, null);
    }

    public MagnoliaCriteriaWithLimitImpl(String path, QueryManager queryManager, String itemType, Class< ? > classType)
    {
        super(path, queryManager, itemType, classType);
    }

    @SuppressWarnings("unchecked")
    @Override
    public Collection< ? > list() throws JCRQueryException
    {
        if (queryManager == null)
            throw new IllegalStateException("QueryManager cannot be null");

        log.debug("creating statement...");
        long start = System.currentTimeMillis();

        JCRMagnoliaCriteriaQueryTranslator translator = new JCRMagnoliaCriteriaQueryTranslator(this);
        XPathSelect statement = new XPathSelect();
        statement.setRoot(this.path);
        statement.setPredicate(translator.getPredicate());
        statement.setOrderByClause(translator.getOrderBy());
        String stmt = statement.toStatementString();

        long stop = System.currentTimeMillis();
        log.debug("statement created in {} milliseconds", stop - start);

        try
        {
            Class< ? extends QueryManager> managerClass = queryManager.getClass();
            java.lang.reflect.Field field = managerClass.getDeclaredField(QUERY_MANAGER);
            field.setAccessible(true);

            javax.jcr.query.QueryManager jcrManager = (javax.jcr.query.QueryManager) field.get(queryManager);
            org.apache.jackrabbit.core.query.QueryImpl q = (org.apache.jackrabbit.core.query.QueryImpl) jcrManager
                .createQuery(stmt, Query.XPATH);

            if (getMaxResults() <= 0)
            {
                String msg = "maxResults value [" + getMaxResults() + "] cannot be less than or equal to 0";
                log.error(msg);
                throw new InvalidQueryException(msg);
            }
            q.setLimit(getMaxResults());

            if (getFirstResult() < 0)
            {
                String msg = "firstResult value [" + getFirstResult() + "] cannot be less than 0";
                log.error(msg);
                throw new InvalidQueryException(msg);
            }
            q.setOffset(getFirstResult());
            log.debug("offset is {} limit is {}", new Object[]{getFirstResult(), getMaxResults() });

            log.debug("executing Query {}", q.getStatement());
            start = System.currentTimeMillis();
            javax.jcr.query.QueryResult result = q.execute();

            stop = System.currentTimeMillis();
            log.debug("Query executed in {} milliseconds", stop - start);

            Collection retVal;
            field = managerClass.getDeclaredField(HIERARCHY_MANAGER);
            field.setAccessible(true);
            HierarchyManager hm = (HierarchyManager) field.get(queryManager);

            log.debug("getting content...");
            start = System.currentTimeMillis();
            QueryResultImpl filteredResult = new QueryResultImpl(result, hm);
            retVal = filteredResult.getContent(itemType);

            stop = System.currentTimeMillis();

            if (Content.class.isAssignableFrom(classType))
            {
                log.debug("returning a Collection holding objects of type {} ", Content.class.getName());
                retVal = retVal == null ? new ArrayList<Content>() : retVal;
                log.debug("got {} node(s) in {} ms", new Object[]{retVal.size(), stop - start });

                return retVal;
            }
            else
            {
                log.debug("returning a Collection holding objects of type {} ", classType.getName());
                Iterator iter = retVal.iterator();
                List list = new ArrayList();
                while (iter.hasNext())
                {
                    Content node = (Content) iter.next();
                    try
                    {
                        list.add(classType.cast(Content2BeanUtil.toBean(node, true, classType)));
                    }
                    catch (Content2BeanException e)
                    {
                        log.error(e.getMessage());
                        throw new JCRQueryException(e);
                    }
                }
                log.debug("got {} node(s) in {} ms", new Object[]{list.size(), stop - start });
                return list;
            }
        }
        catch (InvalidQueryException e)
        {
            log.error(e.getMessage());
            throw new JCRQueryException(e);
        }
        catch (RepositoryException e)
        {
            log.error(e.getMessage());
            throw new JCRQueryException(e);
        }
        catch (SecurityException e)
        {
            log.error(e.getMessage());
            throw new JCRQueryException(e);
        }
        catch (NoSuchFieldException e)
        {
            log.error(e.getMessage());
            throw new JCRQueryException(e);
        }
        catch (IllegalArgumentException e)
        {
            log.error(e.getMessage());
            throw new JCRQueryException(e);
        }
        catch (IllegalAccessException e)
        {
            log.error(e.getMessage());
            throw new JCRQueryException(e);
        }
    }

    /**
     * This is {@link QueryResultImpl} slightly modified
     */
    protected static final class QueryResultImpl implements QueryResult
    {

        /**
         * Unfiltered result object
         */
        private javax.jcr.query.QueryResult result;

        private AccessManager accessManager;

        private HierarchyManager hm;

        private Map<String, String> dirtyHandles = new Hashtable<String, String>();

        protected QueryResultImpl(javax.jcr.query.QueryResult result, HierarchyManager hm)
        {
            this.result = result;
            this.hm = hm;
            this.accessManager = hm.getAccessManager();
        }

        public AccessManager getAccessManager()
        {
            return accessManager;
        }

        public javax.jcr.query.QueryResult getJcrResult()
        {
            return result;
        }

        /**
         * Build required result objects
         */
        private void build(String nodeType, Collection<DefaultContent> collection) throws RepositoryException
        {
            NodeIterator nodeIterator = this.result.getNodes();
            while (nodeIterator.hasNext())
            {
                Node node = nodeIterator.nextNode();
                try
                {
                    build(node, nodeType, collection);
                }
                catch (RepositoryException re)
                {
                    log.error("{} caught while iterating on query results: {}", re.getClass().getName(), re
                        .getMessage());
                    if (log.isDebugEnabled())
                    {
                        log.debug(re.getClass().getName()
                            + " caught while iterating on query results: "
                            + re.getMessage(), re);
                    }
                }
            }
        }

        /**
         * Build required result objects
         */
        private void build(Node node, String nodeType, Collection<DefaultContent> collection)
            throws RepositoryException
        {
            /**
             * All custom node types
             */
            if ((node.isNodeType(nodeType) || StringUtils.isEmpty(nodeType)) && !node.isNodeType(ItemType.NT_RESOURCE))
            {
                if (this.dirtyHandles.get(node.getPath()) == null)
                {
                    boolean isAllowed = this.accessManager.isGranted(
                        Path.getAbsolutePath(node.getPath()),
                        Permission.READ);
                    if (isAllowed)
                    {
                        collection.add(new DefaultContent(node, this.hm));
                        this.dirtyHandles.put(node.getPath(), StringUtils.EMPTY);
                    }
                }
                return;
            }
            if (node.getDepth() > 0)
            {
                this.build(node.getParent(), nodeType, collection);
            }
        }

        /**
         * @see info.magnolia.cms.core.search.QueryResult#getContent()
         */
        public Collection<DefaultContent> getContent()
        {
            return getContent(ItemType.CONTENT.getSystemName());
        }

        /**
         * @see info.magnolia.cms.core.search.QueryResult#getContent(java.lang.String)
         */
        public Collection<DefaultContent> getContent(String nodeType)
        {
            Collection<DefaultContent> resultSet = new ArrayList<DefaultContent>();
            try
            {
                this.build(nodeType, resultSet);
            }
            catch (RepositoryException re)
            {
                log.error(re.getMessage());
            }
            return resultSet;
        }
    }
}